"""
Copyright (c) 2013-2018 University of Oxford
"""
__contrib__ = {
    "Martin Craig" : "Current maintainer",
    "Benjamin Irving" : "Original author",
    "Michael Chappell" : "",
    "Paula Croal" : "",
}

__acknowledge__ = {
    "Julia Schnabel" : "",
    "Sir Mike Brady" : "",
}
