"""
Quantiphyse - Widget which displays list of all data loaded

Copyright (c) 2013-2018 University of Oxford
"""

from .widget import OverviewWidget

QP_MANIFEST = {
    "widgets" : [OverviewWidget],
}
